#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

# flake8: noqa
from .algorithm import AlgorithmType, Algorithm
from .user_defined import DaskBFE, DaskIsland
from .protocols import IslandProtocol, ProblemSingleObjective
from .util import (
    CalibrationResult,
    CalibrationMode,
    Island,
    check_ranges,
    list_to_slice,
    read_data,
    read_datacubes,
    list_to_3d_slice,
)
from ..pipelines.processor import ResultType
from .archipelago import MyArchipelago
from .calibration import Calibration, CalibrationMode
from .fitness import sum_of_abs_residuals, sum_of_squared_residuals
from .fitting import ModelFitting
